use anyhow::{Context, Result};
use nalgebra::{Point2, Vector2};

use crate::{handle::SpriteHandle, Content, ContentBuildContext, SectionEdge};

pub(crate) mod syntax {
	use crate::{sprite::syntax::SectionEdge, Content, ContentBuildContext};
	use anyhow::{bail, Context, Result};
	use nalgebra::{Point2, Vector2};
	use serde::Deserialize;
	// Raw serde syntax structs.
	// These are never seen by code outside this crate.

	#[derive(Debug, Deserialize)]
	pub struct Ui {
		pub landed: UiLanded,
	}

	#[derive(Debug, Deserialize)]
	pub struct UiLanded {
		pub frame: UiSprite,
		pub landscape: UiSprite,
		pub button: UiSprite,
	}

	#[derive(Debug, Deserialize)]
	pub struct UiSprite {
		pub sprite: Option<String>,
		pub pos: [f32; 2],
		pub dim: [f32; 2],
		pub loc_div: Option<f32>,
		pub mask: Option<String>,
		pub on_mouse_enter: Option<EdgeSpec>,
		pub on_mouse_leave: Option<EdgeSpec>,
	}

	#[derive(Debug, Deserialize)]
	pub struct EdgeSpec {
		pub edge: SectionEdge,
		pub duration: f32,
	}

	impl UiSprite {
		pub fn build(
			self,
			build_context: &ContentBuildContext,
			ct: &Content,

			// If true, this ui sprite will be positioned relative to another
			is_child: bool,

			// If true, fail if self.sprite is missing.
			// If false, fail if self.sprite exists.
			// This is false for sprites that may change---for example, planet landscapes
			should_have_sprite: bool,
		) -> Result<super::UiSpriteConfig> {
			let sprite = {
				if should_have_sprite {
					if self.sprite.is_none() {
						bail!("no sprite given, but expected a value")
					}
					match ct.sprite_index.get(self.sprite.as_ref().unwrap()) {
						None => bail!("ui sprite `{}` doesn't exist", self.sprite.unwrap()),
						Some(t) => Some(*t),
					}
				} else {
					if self.sprite.is_some() {
						bail!("got a sprite, but didn't expect one")
					}
					None
				}
			};

			let mask = if let Some(mask) = self.mask {
				Some(match ct.sprite_index.get(&mask) {
					None => bail!("mask `{}` doesn't exist", mask),
					Some(t) => *t,
				})
			} else {
				None
			};

			let on_mouse_enter = {
				if let Some(x) = self.on_mouse_enter {
					if sprite.is_none() {
						bail!("got `on_mouse_enter` on a ui element with no fixed sprite")
					}

					Some(
						x.edge
							.resolve_as_edge(sprite.unwrap(), build_context, x.duration)
							.with_context(|| format!("failed to resolve mouse enter edge"))?,
					)
				} else {
					None
				}
			};

			let on_mouse_leave = {
				if let Some(x) = self.on_mouse_leave {
					if sprite.is_none() {
						bail!("got `on_mouse_leave` on a ui element with no fixed sprite")
					}

					Some(
						x.edge
							.resolve_as_edge(sprite.unwrap(), build_context, x.duration)
							.with_context(|| format!("failed to resolve mouse leave edge"))?,
					)
				} else {
					None
				}
			};

			let d = self.loc_div.unwrap_or(1.0);

			let rect = {
				if is_child {
					super::UiSpriteRect::Relative {
						pos: Point2::new(self.pos[0], self.pos[1]) / d,
						dim: Vector2::new(self.dim[0], self.dim[1]) / d,
					}
				} else {
					super::UiSpriteRect::Absolute {
						pos: Point2::new(self.pos[0], self.pos[1]) / d,
						dim: Vector2::new(self.dim[0], self.dim[1]) / d,
					}
				}
			};

			return Ok(super::UiSpriteConfig {
				sprite,
				mask,
				on_mouse_enter,
				on_mouse_leave,
				rect,
			});
		}
	}
}

/// UI Configuration
#[derive(Debug, Clone)]
pub struct Ui {
	/// Landed interface frame
	pub landed_frame: UiSpriteConfig,

	/// Landed interface image
	pub landed_landscape: UiSpriteConfig,

	/// Test button
	pub landed_button: UiSpriteConfig,
}

/// A UI sprite's position
#[derive(Debug, Clone)]
pub enum UiSpriteRect {
	/// Positioning relative to a parent sprite
	Relative {
		// Note that both pos and dim include transparent pixels,
		// of this sprite AND its parent.

		// We use the top left corner here because that's how inkscape
		// positions its objects. This makes it very easy to compute position.
		// TODO: maybe add anchors here too?
		/// The position of this sprite's northeast corner, relative to its parent's NE corner.
		/// Positive X is right, positive Y is down.
		pos: Point2<f32>,

		/// This sprite's w and h, as a fraction of the parent's width and height.
		dim: Vector2<f32>,
	},

	/// Absolute positioning
	Absolute {
		/// The position of the center of this sprite, in logical pixels,
		/// with 0, 0 at the center of the screen
		pos: Point2<f32>,

		/// This sprite's w and h, in logical pixels.
		dim: Vector2<f32>,
	},
}

impl UiSpriteRect {
	/// Get this rectangle's position
	pub fn get_pos(&self) -> &Point2<f32> {
		match self {
			Self::Relative { pos, .. } => pos,
			Self::Absolute { pos, .. } => pos,
		}
	}

	/// Get this rectangle's dimensions
	pub fn get_dim(&self) -> &Vector2<f32> {
		match self {
			Self::Relative { dim, .. } => dim,
			Self::Absolute { dim, .. } => dim,
		}
	}
}

/// A single UI sprite instance
#[derive(Debug, Clone)]
pub struct UiSpriteConfig {
	/// The sprite to show
	pub sprite: Option<SpriteHandle>,

	/// The mask to use
	pub mask: Option<SpriteHandle>,

	/// This sprite's position and size
	pub rect: UiSpriteRect,

	/// Animation edge to take when mouse enters this sprite
	pub on_mouse_enter: Option<SectionEdge>,

	/// Animation edge to take when mouse leaves this sprite
	pub on_mouse_leave: Option<SectionEdge>,
}

impl crate::Build for Ui {
	type InputSyntaxType = syntax::Ui;

	fn build(
		ui: Self::InputSyntaxType,
		build_context: &mut ContentBuildContext,
		ct: &mut Content,
	) -> Result<()> {
		ct.ui = Some(Ui {
			landed_frame: ui
				.landed
				.frame
				.build(build_context, ct, false, true)
				.with_context(|| format!("in ui config (frame)"))?,
			landed_landscape: ui
				.landed
				.landscape
				.build(build_context, ct, true, false)
				.with_context(|| format!("in ui config (image)"))?,
			landed_button: ui
				.landed
				.button
				.build(build_context, ct, true, true)
				.with_context(|| format!("in ui config (button)"))?,
		});

		return Ok(());
	}
}
