mod game;

use anyhow::{bail, Result};
use clap::Parser;
use galactica_content::{Content, SystemHandle};
use galactica_playeragent::{PlayerAgent, PlayerStatus};
use galactica_render::RenderInput;
use galactica_system::{
	data::ShipState,
	phys::{PhysImage, PhysSimShipHandle},
};
use galactica_util::constants::ASSET_CACHE;
use log::LevelFilter;
use log4rs::{
	append::console::ConsoleAppender,
	config::{Appender, Logger, Root},
	encode::pattern::PatternEncoder,
	Config,
};
use nalgebra::Vector2;
use std::{
	fs,
	path::{Path, PathBuf},
	time::Instant,
};
use winit::{
	event::{Event, KeyboardInput, WindowEvent},
	event_loop::{ControlFlow, EventLoop},
	window::WindowBuilder,
};

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
	/// Run the asset packer
	#[arg(long, action, group = "run")]
	pack: bool,

	/// Check content in the default directory
	#[arg(long, action, group = "run")]
	check_content: bool,
}

mod cli;

fn set_up_logger(_args: &Args) -> log4rs::Handle {
	// h: highlight with level color
	// :<0.128 right align, min 0, max 128
	let stdout = ConsoleAppender::builder()
		.encoder(Box::new(PatternEncoder::new(
			"{d(%H:%M:%S)} | {({h({l})}):5.5} | \x1b[0;34m{M}\x1b[0m — {m:<0.128}{n}",
		)))
		.build();

	let config = Config::builder()
		.appender(Appender::builder().build("stdout", Box::new(stdout)))
		.logger(Logger::builder().build("wgpu_core", LevelFilter::Warn))
		.logger(Logger::builder().build("wgpu_hal", LevelFilter::Warn))
		.logger(Logger::builder().build("naga", LevelFilter::Warn))
		.logger(Logger::builder().build("cosmic_text", LevelFilter::Warn))
		.logger(Logger::builder().build("winit", LevelFilter::Warn))
		.logger(Logger::builder().build("mio", LevelFilter::Warn))
		.build(Root::builder().appender("stdout").build(LevelFilter::Trace))
		.unwrap();

	log4rs::init_config(config).unwrap()
}

fn main() -> Result<()> {
	let args = Args::parse();

	let _ = set_up_logger(&args);

	let cache_dir = Path::new(ASSET_CACHE);
	fs::create_dir_all(cache_dir)?;

	if args.pack {
		cli::packer::run(Path::new("./assets/render"), cache_dir)?;
	} else if args.check_content {
		let atlas_index = cache_dir.join("spriteatlas.toml");
		let _ = Content::load_dir(
			PathBuf::from("./content"),
			PathBuf::from("./assets"),
			atlas_index.clone(),
		)?;
		println!("Content was loaded without errors.")
	} else {
		// Run game
		if let Err(err) = try_main() {
			eprintln!("Galactica failed with an error");

			let mut i = 0;
			for e in err.chain().rev() {
				eprintln!("{i:02}: {}", e);
				i += 1;
			}
			std::process::exit(1);
		}
	}

	return Ok(());
}

fn try_main() -> Result<()> {
	let cache_dir = Path::new(ASSET_CACHE);
	fs::create_dir_all(cache_dir)?;

	let atlas_index = cache_dir.join("spriteatlas.toml");

	if !atlas_index.exists() {
		bail!("Could not find sprite atlas!");
	}

	// TODO: pretty error if missing (also in cli)
	let content = Content::load_dir(
		PathBuf::from("./content"),
		PathBuf::from("./assets"),
		atlas_index,
	)?;

	let event_loop = EventLoop::new();
	let window = WindowBuilder::new().build(&event_loop).unwrap();
	let mut gpu = pollster::block_on(galactica_render::GPUState::new(window, &content))?;
	gpu.init(&content);

	// TODO: don't clone content
	let mut game = game::Game::new(content.clone());
	let p = game.make_player();

	let mut player = PlayerAgent::new(p.0);
	player.set_camera_aspect(
		gpu.window().inner_size().width as f32 / gpu.window().inner_size().height as f32,
	);

	let mut phys_img = PhysImage::new();
	let mut last_run = Instant::now();

	event_loop.run(move |event, _, control_flow| {
		match event {
			Event::RedrawRequested(window_id) if window_id == gpu.window().id() => {
				let render_input = RenderInput {
					camera_pos: player.camera.pos,
					camera_zoom: player.camera.zoom,
					current_time: game.get_current_time(),
					ct: &content,
					phys_img: &phys_img,
					player: &player,
					time_since_last_run: last_run.elapsed().as_secs_f32(),
					current_system: SystemHandle { index: 0 },
					timing: game.get_timing().clone(),
				};
				last_run = Instant::now();

				match gpu.render(render_input) {
					Ok(_) => {}
					Err(wgpu::SurfaceError::Lost) => gpu.resize(&content),
					Err(wgpu::SurfaceError::OutOfMemory) => *control_flow = ControlFlow::Exit,
					// All other errors (Outdated, Timeout) should be resolved by the next frame
					Err(e) => eprintln!("{:?}", e),
				}
			}

			Event::MainEventsCleared => {
				game.update_player_controls(&mut player);
				game.step(&phys_img);
				game.update_image(&mut phys_img);

				// TODO: clean up
				let player_status = {
					let pos = {
						let o = phys_img.get_ship(&PhysSimShipHandle(player.ship.unwrap()));
						if let Some(o) = o {
							match o.ship.get_data().get_state() {
								ShipState::Landing { .. }
								| ShipState::UnLanding { .. }
								| ShipState::Collapsing { .. }
								| ShipState::Flying { .. } => Some(*o.rigidbody.translation()),

								ShipState::Landed { target } => {
									let b = content.get_system_object(*target);
									Some(Vector2::new(b.pos.x, b.pos.y))
								}

								ShipState::Dead => None,
							}
						} else {
							None
						}
					};

					PlayerStatus { pos }
				};

				// This must be updated BEFORE rendering!
				player.step(&content, player_status);

				player.input.clear_inputs();
				gpu.window().request_redraw();
			}

			Event::WindowEvent {
				ref event,
				window_id,
			} if window_id == gpu.window().id() => match event {
				WindowEvent::Focused(_state) => {
					//game.set_paused(!state);
				}
				WindowEvent::CloseRequested => *control_flow = ControlFlow::Exit,
				WindowEvent::KeyboardInput {
					input:
						KeyboardInput {
							state,
							virtual_keycode: Some(key),
							..
						},
					..
				} => {
					player.input.process_key(state, key);
				}
				WindowEvent::CursorMoved { position, .. } => {
					player.input.process_mouse(position);
				}
				WindowEvent::MouseInput { state, button, .. } => {
					player.input.process_click(state, button);
				}
				WindowEvent::MouseWheel { delta, phase, .. } => {
					player.input.process_scroll(delta, phase);
				}
				WindowEvent::Resized(_) => {
					gpu.resize(&content);
					player.set_camera_aspect(
						gpu.window().inner_size().width as f32
							/ gpu.window().inner_size().height as f32,
					);
				}
				WindowEvent::ScaleFactorChanged { .. } => {
					gpu.resize(&content);
					player.set_camera_aspect(
						gpu.window().inner_size().width as f32
							/ gpu.window().inner_size().height as f32,
					);
				}
				_ => {}
			},
			_ => {}
		}
	});
}
