use galactica_content::{Content, SystemObject, SystemObjectHandle};
use galactica_system::{data::ShipState, phys::PhysSimShipHandle};
use glyphon::{cosmic_text::Align, Attrs, Color, Metrics, TextArea, Weight};
use nalgebra::{Point2, Vector2};

use super::util::{SpriteRect, UiSprite, UiTextArea};
use crate::{RenderInput, RenderState};

pub(super) struct Planet {
	// UI elements
	planet_desc: UiTextArea,
	planet_name: UiTextArea,
	frame: UiSprite,
	landscape: UiSprite,
	button: UiSprite,

	/// What object we're displaying currently.
	/// Whenever this changes, we need to reflow text.
	current_object: Option<SystemObjectHandle>,
}

// TODO: animate in/out

impl Planet {
	pub fn new(ct: &Content, state: &mut RenderState) -> Self {
		let frame = UiSprite::from(ct, &ct.get_ui().landed_frame);
		let button = UiSprite::from(ct, &ct.get_ui().landed_button);
		let landscape = UiSprite::from_with_sprite(
			ct,
			&ct.get_ui().landed_landscape,
			ct.get_sprite_handle("ui::landscape::test"),
		);

		let s = Self {
			// height of element in logical pixels
			current_object: None,

			planet_desc: UiTextArea::new(
				ct,
				state,
				frame.get_sprite(),
				Point2::new(0.0, 0.0),
				frame.get_height(),
				SpriteRect {
					pos: Point2::new(25.831, 284.883) / 512.0,
					dim: Vector2::new(433.140, 97.220) / 512.0,
				},
				Metrics::new(16.0, 18.0),
				Color::rgb(255, 255, 255),
				Align::Left,
			),

			planet_name: UiTextArea::new(
				ct,
				state,
				frame.get_sprite(),
				Point2::new(0.0, 0.0),
				frame.get_height(),
				SpriteRect {
					pos: Point2::new(165.506, 82.0) / 512.0,
					dim: Vector2::new(74.883, 17.0) / 512.0,
				},
				Metrics::new(19.0, 19.0),
				Color::rgb(255, 255, 255),
				Align::Center,
			),

			frame,
			landscape,
			button,
		};

		return s;
	}
}

impl Planet {
	fn reflow(&mut self, planet: &SystemObject, state: &mut RenderState) {
		self.planet_desc.set_text(
			state,
			&planet.desc,
			Attrs::new()
				.weight(Weight::NORMAL)
				.family(glyphon::Family::SansSerif),
		);

		self.planet_name.set_text(
			state,
			&planet.name,
			Attrs::new()
				.weight(Weight::BOLD)
				.family(glyphon::Family::Serif),
		);
		self.current_object = Some(planet.handle);
	}

	pub fn draw(&mut self, input: &RenderInput, state: &mut RenderState) {
		// Get required data
		let ship_handle = input.player.ship.unwrap();
		let ship_data = &input
			.phys_img
			.get_ship(&PhysSimShipHandle(ship_handle))
			.unwrap()
			.ship;
		let planet_handle = match ship_data.get_data().get_state() {
			ShipState::Landed { target } => *target,
			_ => unreachable!("tried to draw planet interface while not landed!"),
		};
		let planet = input.ct.get_system_object(planet_handle);

		// Reconfigure for new planet if necessary
		if self
			.current_object
			.map(|x| x != planet_handle)
			.unwrap_or(true)
		{
			self.reflow(planet, state);
		}

		self.button.step(input, state);
		self.landscape.step(input, state);
		self.frame.step(input, state);

		// Draw elements
		self.button
			.push_to_buffer(input, state, Some(self.frame.get_rect(input)));
		self.landscape
			.push_to_buffer(input, state, Some(self.frame.get_rect(input)));
		self.frame.push_to_buffer(input, state, None);
	}

	pub fn get_textarea(&self, input: &RenderInput, state: &RenderState) -> [TextArea; 2] {
		[
			self.planet_desc.get_textarea(input, state),
			self.planet_name.get_textarea(input, state),
		]
	}
}
