use galactica_content::{Content, SectionEdge, SpriteAutomaton, SpriteHandle, UiSpriteConfig};
use galactica_util::to_radians;
use nalgebra::{Point2, Vector2};

use super::SpriteRect;
use crate::{vertexbuffer::types::UiInstance, PositionAnchor, RenderInput, RenderState};

pub struct UiSprite {
	pub anim: SpriteAutomaton,
	mask: Option<SpriteHandle>,

	rect: SpriteRect,
	has_mouse: bool,

	on_mouse_enter: Option<SectionEdge>,
	on_mouse_leave: Option<SectionEdge>,
}

impl UiSprite {
	pub fn from(ct: &Content, ui: &UiSpriteConfig) -> Self {
		Self::from_with_sprite(ct, ui, ui.sprite.unwrap())
	}

	pub fn from_with_sprite(ct: &Content, ui: &UiSpriteConfig, sprite: SpriteHandle) -> Self {
		if ui.sprite.is_none() {
			unreachable!("called `UiSprite.from()` on a UiSprite with a None sprite!")
		}

		return Self {
			anim: SpriteAutomaton::new(ct, sprite),
			mask: ui.mask,
			rect: SpriteRect {
				pos: *ui.rect.get_pos(),
				dim: *ui.rect.get_dim(),
			},
			has_mouse: false,
			on_mouse_enter: ui.on_mouse_enter,
			on_mouse_leave: ui.on_mouse_leave,
		};
	}

	pub fn step(&mut self, input: &RenderInput, state: &RenderState) {
		if self.contains_mouse(input, state, Some(self.get_rect(input)))
			&& !self.has_mouse
			&& self.on_mouse_enter.is_some()
		{
			self.has_mouse = true;
			self.anim.jump_to(input.ct, self.on_mouse_enter.unwrap())
		}

		if !self.contains_mouse(input, state, Some(self.get_rect(input)))
			&& self.has_mouse
			&& self.on_mouse_leave.is_some()
		{
			self.has_mouse = false;
			self.anim.jump_to(input.ct, self.on_mouse_leave.unwrap())
		}

		self.anim.step(input.ct, input.time_since_last_run);
	}

	pub fn contains_mouse(
		&self,
		input: &RenderInput,
		state: &RenderState,
		parent: Option<SpriteRect>,
	) -> bool {
		let rect = self.get_relative_rect(input, parent);

		let fac = state.window.scale_factor() as f32;
		let window_size = Vector2::new(
			state.window_size.width as f32 / fac,
			state.window_size.height as f32 / fac,
		);

		let pos = input.player.input.get_mouse_pos();
		let mouse_pos = Vector2::new(
			pos.x / fac - window_size.x / 2.0,
			window_size.y / 2.0 - pos.y / fac,
		);

		let ne = rect.ne_corner();
		let sw = rect.sw_corner();
		return (mouse_pos.y < ne.y && mouse_pos.y > sw.y)
			&& (mouse_pos.x > ne.x && mouse_pos.x < sw.x);
	}

	pub fn get_rect(&self, input: &RenderInput) -> SpriteRect {
		let pos = Point2::new(self.rect.pos.x, self.rect.pos.y);
		let dim = Vector2::new(
			self.rect.dim.y * input.ct.get_sprite(self.anim.get_sprite()).aspect,
			self.rect.dim.y,
		);

		return SpriteRect { dim, pos };
	}

	pub fn get_relative_rect(&self, input: &RenderInput, parent: Option<SpriteRect>) -> SpriteRect {
		if let Some(parent) = parent {
			let zero = Point2::new(
				parent.pos.x - (parent.dim.x / 2.0),
				parent.pos.y + (parent.dim.y / 2.0),
			);

			let mut pos = zero
				+ Vector2::new(
					self.rect.pos.x * parent.dim.x,
					-self.rect.pos.y * parent.dim.y,
				);
			let dim = Vector2::new(
				self.rect.dim.x * parent.dim.x,
				self.rect.dim.y * parent.dim.y,
			);

			pos += Vector2::new(dim.x, -dim.y) / 2.0;

			return SpriteRect { dim, pos };
		} else {
			return self.get_rect(input);
		}
	}

	/// Add this image to the gpu sprite buffer
	pub fn push_to_buffer(
		&self,
		input: &RenderInput,
		state: &mut RenderState,
		parent: Option<SpriteRect>,
	) {
		let rect = self.get_relative_rect(input, parent.clone());

		// TODO: use both dimensions,
		// not just height
		let anim_state = self.anim.get_texture_idx();
		state.push_ui_buffer(UiInstance {
			anchor: PositionAnchor::CC.to_int(),
			position: rect.pos.into(),
			angle: to_radians(90.0),
			size: rect.dim.y,
			color: [1.0, 1.0, 1.0, 1.0],
			texture_index: anim_state.texture_index(),
			texture_fade: anim_state.fade,
			mask_index: self
				.mask
				.map(|x| {
					let sprite = input.ct.get_sprite(x);
					let texture_b = sprite.get_first_frame(); // ANIMATE
					[1, texture_b]
				})
				.unwrap_or([0, 0]),
		});
	}

	pub fn get_sprite(&self) -> SpriteHandle {
		self.anim.get_sprite()
	}

	pub fn get_height(&self) -> f32 {
		self.rect.dim.y
	}
}
