use galactica_content::{Content, SpriteHandle};
use glyphon::{cosmic_text::Align, Attrs, Buffer, Color, Metrics, Shaping, TextArea, TextBounds};
use nalgebra::{Point2, Vector2};

use super::SpriteRect;
use crate::{RenderInput, RenderState};

/// Represents a text area inside a sprite.
pub(crate) struct UiTextArea {
	/// Parent sprite
	sprite: SpriteHandle,

	/// Position of parent sprite's center, in logical pixels,
	/// with 0, 0 at the center of the screen
	sprite_position: Point2<f32>,

	/// Height of parent sprite, in logical pixels
	sprite_size: f32,

	/// Bounds of text area
	rect: SpriteRect,

	/// Text buffer
	buffer: Buffer,

	/// Text color
	color: Color,

	/// Text alignment
	align: Align,
}

impl UiTextArea {
	pub fn new(
		ct: &Content,
		state: &mut RenderState,
		sprite: SpriteHandle,
		sprite_position: Point2<f32>,
		sprite_size: f32,
		rect: SpriteRect,
		text_metrics: Metrics,
		color: Color,
		align: Align,
	) -> Self {
		let mut s = Self {
			buffer: Buffer::new(&mut state.text_font_system, text_metrics),
			sprite_size: f32::NAN,
			sprite,
			sprite_position,
			rect,
			align,
			color,
		};
		s.set_size(ct, state, sprite_size);
		return s;
	}

	pub fn set_size(&mut self, ct: &Content, state: &mut RenderState, sprite_size: f32) {
		self.sprite_size = sprite_size;
		self.buffer.set_size(
			&mut state.text_font_system,
			(self.rect.dim.x * self.sprite_size) * state.window.scale_factor() as f32,
			(self.rect.dim.y * self.sprite_size * ct.get_sprite(self.sprite).aspect)
				* state.window.scale_factor() as f32,
		);
	}

	pub fn set_text(&mut self, state: &mut RenderState, text: &str, attrs: Attrs) {
		self.buffer
			.set_text(&mut state.text_font_system, text, attrs, Shaping::Advanced);

		for l in &mut self.buffer.lines {
			l.set_align(Some(self.align));
		}
		self.buffer.shape_until_scroll(&mut state.text_font_system);
	}

	pub fn get_textarea(&self, input: &RenderInput, state: &RenderState) -> TextArea {
		let h = self.sprite_size;
		let w = input.ct.get_sprite(self.sprite).aspect * h;

		// Glypon works with physical pixels, so we must convert
		let fac = state.window.scale_factor() as f32;

		// All the units below are in logical pixels
		let zero = Vector2::new(
			(state.window_size.width as f32 / (2.0 * fac)) - (w / 2.0) + self.sprite_position.x,
			(state.window_size.height as f32 / (2.0 * fac)) - (h / 2.0) - self.sprite_position.y,
		);
		let corner_ne = zero + Vector2::new(self.rect.pos.x * w, self.rect.pos.y * h);
		let corner_sw = corner_ne + Vector2::new(self.rect.dim.x * w, self.rect.dim.y * h);

		TextArea {
			buffer: &self.buffer,
			top: corner_ne.y * fac,
			left: corner_ne.x * fac,
			scale: 1.0,
			bounds: TextBounds {
				top: (corner_ne.y * fac) as i32,
				bottom: (corner_sw.y * fac) as i32,
				left: (corner_ne.x * fac) as i32,
				right: (corner_sw.x * fac) as i32,
			},
			default_color: self.color,
		}
	}
}
